.plot-zoom {
    position: absolute;
    border: none;
    background-color: transparent;
    bottom: 0;
    right: 0;

}
.full-screen {
    position: fixed;
    height: 98vh !important;
    width: 98vw !important;
    left: 0;
    top: 0;
    z-index: 9999;
    overflow: hidden;
}